package com.jakubu9333.grammarManager.service;

import com.jakubu9333.grammarManager.dao.FileMetaDataRepo;
import com.jakubu9333.grammarManager.dao.XmlDataDao;
import com.jakubu9333.grammarManager.dao.XmlDataRepo;
import com.jakubu9333.grammarManager.model.FileHolder;
import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.RuleModel;
import edu.cmu.sphinx.jsgf.JSGFGrammarException;
import edu.cmu.sphinx.jsgf.JSGFGrammarParseException;
import edu.cmu.sphinx.jsgf.JSGFRuleGrammar;
import edu.cmu.sphinx.jsgf.JSGFRuleGrammarFactory;
import edu.cmu.sphinx.jsgf.JSGFRuleGrammarManager;
import edu.cmu.sphinx.jsgf.parser.JSGFParser;
import edu.cmu.sphinx.jsgf.parser.TokenMgrError;
import edu.cmu.sphinx.jsgf.rule.JSGFRule;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleAlternatives;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleCount;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleName;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleSequence;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleTag;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleToken;
import org.javatuples.Pair;
import org.jvoicexml.processor.grammar.RuleAlternatives;
import org.jvoicexml.processor.grammar.RuleComponent;
import org.jvoicexml.processor.grammar.RuleCount;
import org.jvoicexml.processor.grammar.RuleParse;
import org.jvoicexml.processor.grammar.RuleReference;
import org.jvoicexml.processor.grammar.RuleSequence;
import org.jvoicexml.processor.grammar.RuleSpecial;
import org.jvoicexml.processor.grammar.RuleTag;
import org.jvoicexml.processor.grammar.RuleToken;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.server.ResponseStatusException;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import javax.transaction.Transactional;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import java.io.IOException;
import java.io.StringReader;
import java.net.URISyntaxException;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * @author Jakub Uhlarik
 */
@Service
public class FileService {

    private final FileMetaDataRepo fileMetaDataRepo;

    private final GrammarService grammarService;
    private final XmlDataRepo xmlDataRepo;
    private final XmlDataDao xmlDataDao;

    @Autowired
    public FileService(FileMetaDataRepo fileMetaDataRepo, GrammarService grammarService, XmlDataRepo xmlDataRepo, XmlDataDao xmlDataDao) {
        this.xmlDataDao = xmlDataDao;
        this.fileMetaDataRepo = fileMetaDataRepo;
        this.grammarService = grammarService;
        this.xmlDataRepo = xmlDataRepo;
    }


    public static class XmlGrammarHandler extends DefaultHandler {
        private Grammar newGrammar;
        private StringBuilder rightSide;

        private RuleModel rule = null;

        private boolean grammar = false;

        @Override
        public void startDocument() {
            newGrammar = new Grammar(null, new HashSet<>(), null);
        }

        @Override
        public void characters(char[] ch, int start, int length) {
            if (rightSide != null) {
                rightSide.append(ch, start, length);
            }
        }

        @Override
        public void startElement(String uri, String lName, String qName, Attributes attr) {
            switch (qName) {
                case "grammar" -> {
                    grammar = true;
                    String grammarUri = attr.getValue("xml:base");
                    if (grammarUri != null) {
                        newGrammar.setUri(grammarUri);
                    }
                    String rootLeft = attr.getValue("root");
                    if (rootLeft != null) {
                        newGrammar.setRootSymbol(rootLeft);
                    }
                    return;
                }
                case "rule" -> {
                    if (!grammar) {
                        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "rule outside grammar");
                    }
                    rule = new RuleModel(newGrammar, attr.getValue("id"), null, false);
                    rightSide = new StringBuilder();
                    return;
                }
                case "item" -> {
                    if (rule == null) {
                        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "item outside rule");
                    }
                }
                case "one-of" -> {
                    if (rule == null) {
                        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "one-of outside rule");
                    }
                }
            }


            if (rule != null) {
                rightSide.append("<").append(qName);

                for (int i = 0; i < attr.getLength(); i++) {
                    rightSide.append(" ").append(attr.getQName(i)).append("=\"").append(attr.getValue(i)).append('"');
                }

                rightSide.append(">");


            }


        }

        @Override
        public void endElement(String uri, String lName, String qName) {

            if ("rule".equals(qName)) {
                rule.setRightSide(rightSide.toString());

                newGrammar.getRules().add(rule);
                rule = null;
                return;
            }
            if (rule != null) {
                rightSide.append("</").append(qName).append(">");
            }
        }
    }

    public FileHolder getFileFromGrammar(Long id) {
        return fileMetaDataRepo.findByGrammarId(id).orElseThrow(() -> new ResponseStatusException(HttpStatus.BAD_REQUEST, "no grammar found"));
    }


    private Grammar parseFile(String file, boolean xml) {
        Grammar grammar;
        if (xml) {
            try {
                SAXParserFactory factory = SAXParserFactory.newInstance();
                SAXParser saxParser = factory.newSAXParser();

                XmlGrammarHandler handler = new XmlGrammarHandler();
                InputSource is = new InputSource(new StringReader(file));
                saxParser.parse(is, handler);
                grammar = handler.newGrammar;

            } catch (ParserConfigurationException | SAXException | IOException e) {
                throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "xml parse error\n" + e.getMessage(), e);
            }

        } else {
            try {
                String xmlString = jsgfParse(file);
                xmlDataDao.write(xmlString);
                grammar = new Grammar();
            } catch (JSGFGrammarParseException | TokenMgrError | IOException | URISyntaxException |
                     JSGFGrammarException e) {
                if (e instanceof JSGFGrammarParseException) {
                    throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "jsgf parse error\n" + ((JSGFGrammarParseException) e).details);
                }
                throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "jsgf parse error\n" + e.getMessage(), e);
            }

        }
        return grammarService.preProcessGrammar(grammar);
    }


    private boolean getFileExtension(String file) {
        Pattern xml = Pattern.compile("^\\W*<\\?xml.*\\?>");
        Matcher mXml = xml.matcher(file);
        if (mXml.find()) {
            return true;
        }
        Pattern jsgf = Pattern.compile("^\\W*#JSGF.*;");
        Matcher mJsgf = jsgf.matcher(file);
        if (mJsgf.find()) {
            return false;
        }
        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "format cant be found from file");
    }

    @Transactional
    public void saveFileToGrammar(Grammar grammar, boolean xml, String file) {
        FileHolder fileHolder = new FileHolder(xml, null, file);
        if (xml) {
            xmlDataDao.write(file);
        }
        fileHolder.setSize(file.length());
        fileMetaDataRepo.save(fileHolder);
        grammar.setFile(fileHolder);
        fileHolder.setGrammar(grammar);

    }


    @Transactional
    public Pair<Grammar, Boolean> addGrammarFromFile(String file, Boolean xml) {
        if (xml == null) {
            xml = getFileExtension(file);

        }
        Grammar grammar = parseFile(file, xml);
        grammar = grammarService.preProcessGrammar(grammar);
        saveFileToGrammar(grammar, xml, file);

        return new Pair<>(grammarService.addGrammar(grammar), true);
    }

    private RuleComponent jsgfToXml(JSGFRule jsgfRule) {
        if (jsgfRule instanceof JSGFRuleAlternatives jsgfRuleAlternatives) {
            if (jsgfRuleAlternatives.getRules().size()==1){
                return jsgfToXml(jsgfRuleAlternatives.getRules().get(0));
            }
            RuleAlternatives ruleAlternatives = new RuleAlternatives();
            List<Float> weights = jsgfRuleAlternatives.getWeights();
            AtomicInteger i = new AtomicInteger();
            jsgfRuleAlternatives.getRules().forEach(p -> {
                if (weights == null) {
                    ruleAlternatives.addAlternative(jsgfToXml(p));
                }else {
                    ruleAlternatives.addAlternative(jsgfToXml(p),weights.get(i.get()));
                    i.getAndIncrement();
                }
            });
            return ruleAlternatives;
        }
        if (jsgfRule instanceof JSGFRuleSequence jsgfRuleSequence) {
            if (jsgfRuleSequence.getRules().size()==1){
                return jsgfToXml(jsgfRuleSequence.getRules().get(0));
            }
            RuleSequence ruleSequence = new RuleSequence();

            jsgfRuleSequence.getRules().forEach(p -> ruleSequence.addElement(jsgfToXml(p)));
            return ruleSequence;
        }
        if (jsgfRule instanceof JSGFRuleCount jsgfRuleCount) {
            int repeat = jsgfRuleCount.getCount();
            RuleComponent ruleComponent = jsgfToXml(jsgfRuleCount.getRule());
            RuleCount ruleCount;
            switch (repeat) {
                case JSGFRuleCount.ZERO_OR_MORE -> ruleCount = new RuleCount(ruleComponent, 0);
                case JSGFRuleCount.OPTIONAL -> ruleCount = new RuleCount(ruleComponent, 0, 1);
                case JSGFRuleCount.ONCE_OR_MORE -> ruleCount = new RuleCount(ruleComponent, 1);
                default -> ruleCount = new RuleCount(ruleComponent, 1, 1);
            }
            return ruleCount;
        }
        if (jsgfRule instanceof JSGFRuleName jsgfRuleName) {
            if (jsgfRuleName.getRuleName().equals("NULL")) {
                return RuleSpecial.NULL;
            }
            if (jsgfRuleName.getRuleName().equals("VOID")) {
                return RuleSpecial.VOID;
            }
            return new RuleParse(new RuleReference(jsgfRuleName.getRuleName()), new RuleToken(" "));
        }
        if (jsgfRule instanceof JSGFRuleToken jsgfRuleToken) {
            return new RuleToken(jsgfRuleToken.getText());
        }
        if (jsgfRule instanceof JSGFRuleTag jsgfRuleTag) {
            return new RuleTag(jsgfRuleTag.getTag());
        }
        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "unknown node");


    }


    private String jsgfParse(String file) throws JSGFGrammarParseException, IOException, URISyntaxException, JSGFGrammarException {
        JSGFRuleGrammar grammarRaw = JSGFParser.newGrammarFromJSGF(new StringReader(file), new JSGFRuleGrammarFactory(new JSGFRuleGrammarManager()));
        Set<RuleModel> rules = new HashSet<>();
        StringBuilder sb = new StringBuilder();
        writeHeader(sb, null, true);
        for (String ruleName : grammarRaw.getRuleNames()) {
            JSGFRule rule = grammarRaw.getRule(ruleName);

            rules.add(new RuleModel(ruleName, rule.toString(), false));
            writeXmlRule(sb, ruleName, jsgfToXml(rule));
        }


        sb.append("</grammar>");

        return sb.toString();
    }

    private void writeXmlRule(StringBuilder sb, String name, RuleComponent ruleComponent) {
        sb.append("<rule id=\"");
        sb.append(name).append("\">\n");
        sb.append(ruleComponent.toStringXML());
        sb.append("</rule>\n");
    }


    private void writeRule(StringBuilder sb, RuleModel rule, boolean xml) {
        sb.append("<").append(rule.getLeftSymbol()).append(">").append(" = ").append(rule.getRightSide()).append(";\n");
    }

    private void writeHeader(StringBuilder sb, Grammar grammar, boolean xml) {
        //jsgf
        if (!xml) {
            sb.append("#JSGF  V1.0;\n");
            sb.append("grammar ").append(grammar.getUri()).append(";\n\n");
            return;
        }
        //xml
        sb.append("<?xml version=\"1.0\"?>\n");
        sb.append("<grammar version=\"1.0\"\n");
        sb.append(" xmlns=\"http://www.w3.org/2001/06/grammar\">\n");
    }

}

