package com.jakubu9333.grammarManager.service;



import com.jakubu9333.grammarManager.dao.GrammarRepository;
import com.jakubu9333.grammarManager.dao.RuleRepository;

import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.RuleModel;
import org.springframework.beans.factory.annotation.Autowired;

import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.server.ResponseStatusException;


import javax.transaction.Transactional;


import java.util.HashSet;
import java.util.List;

import java.util.Objects;
import java.util.Random;
import java.util.Set;

import static java.lang.Math.abs;


/**
 * @author Jakub Uhlarik
 */
@Service
public class GrammarService {

    private final GrammarRepository grammarRepository;
    private final RuleRepository ruleRepository;






    @Autowired
    public GrammarService(GrammarRepository grammarRepository,
                          RuleRepository ruleRepository) {
        this.grammarRepository = grammarRepository;
        this.ruleRepository = ruleRepository;
    }

    public List<Grammar> getGrammars() {
        return grammarRepository.findAll();
    }


    @Transactional
    public void deleteAll() {
        grammarRepository.deleteAll();
    }


    @Transactional
    public void deleteById(Long id) {
        grammarRepository.deleteById(id);
    }

    public Grammar preProcessGrammar(Grammar grammar) {
        grammar.setId(null);
        if (grammar.getUri() == null) {
            grammar.setUri("null");
        }

        String rootRule = grammar.getRootSymbol();
        Set<RuleModel> rules = grammar.getRules();
        if (rootRule != null) {
            //ads root attribute to rule in rules
            rules.forEach(p -> {
                if (Objects.equals(p.getLeftSymbol(), rootRule)) {
                    p.setRoot(true);
                }
                p.setGrammar(grammar);
            });
        }


        return grammar;
    }

    @Transactional
    public Grammar processAddGrammar(Grammar grammar) {
        return addGrammar(preProcessGrammar(grammar));
    }

    @Transactional
    public Grammar addGrammar(Grammar grammar) {
        grammar = grammarRepository.save(grammar);
        Boolean exists_uri = grammarRepository.existsByUriAndIdNot(grammar.getUri(), grammar.getId());
        //uri modification
        if (exists_uri) {
            grammar.setUri(generateUniqueUri(grammar.getUri(), grammar.getId()));
        }
        return grammar;
    }

    public Grammar getGrammarById(Long id) {
        return grammarRepository.findById(id).orElseThrow(() -> new ResponseStatusException(HttpStatus.NOT_FOUND, "grammar id: " + id + " not found"));
    }

    @Transactional
    public Grammar duplicateGrammar(Long id) {
        Grammar grammar = getGrammarById(id);
        Grammar newGrammar = new Grammar();
        newGrammar.setUri(grammar.getUri());
        newGrammar.setRootSymbol(grammar.getRootSymbol());
        Set<RuleModel> rules = grammar.getRules();
        Set<RuleModel> newRules = new HashSet<>();
        for (RuleModel rule : rules) {
            newRules.add(new RuleModel(rule));
        }
        newGrammar.setRules(newRules);
        newGrammar = addGrammar(newGrammar);
        return newGrammar;
    }

    public List<Grammar> getGrammarByRootSymbol(String rootSymbol) {
        return grammarRepository.findAllByRootSymbol(rootSymbol);

    }

    public List<Grammar> getGrammarsByUriStart(String uri) {
        return grammarRepository.findByUriStartingWith(uri);
    }


    private String generateUniqueUri(String uri, Long id) {
        String result = uri + id;
        Random random = new Random(System.currentTimeMillis());
        while (grammarRepository.existsByUriAndIdNot(result, id)) {
            result = uri + id;
            result += "_" + abs(random.nextInt());
        }
        return result;
    }

    @Transactional
    public Grammar changeUri(Long id, Grammar grammar) {
        Grammar oldGrammar = this.getGrammarById(id);
        if (grammar.getUri() != null && grammar.getUri().trim().length() != 0) {
            oldGrammar.setUri(grammar.getUri());

            if (grammarRepository.existsByUriAndIdNot(grammar.getUri(), id)) {
                oldGrammar.setUri(generateUniqueUri(grammar.getUri(), oldGrammar.getId()));
                return oldGrammar;
            }
        }
        return oldGrammar;
    }
}
