package com.jakubu9333.grammarManager.service;

import com.jakubu9333.grammarManager.dao.GrammarRepository;

import com.jakubu9333.grammarManager.dao.RuleRepository;
import com.jakubu9333.grammarManager.model.Grammar;

import com.jakubu9333.grammarManager.model.RuleModel;

import org.javatuples.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.server.ResponseStatusException;

import javax.transaction.Transactional;
import java.time.LocalDateTime;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;


/**
 * @author Jakub Uhlarik
 */
@Service
public class RuleService {

    private final RuleRepository ruleRepository;
    private final GrammarRepository grammarRepository;

    @Autowired
    public RuleService(RuleRepository ruleRepository, GrammarRepository grammarRepository) {
        this.ruleRepository = ruleRepository;
        this.grammarRepository = grammarRepository;
    }

    public List<RuleModel> getAllRules() {
        return ruleRepository.findAll();
    }

    private RuleModel saveRule(RuleModel rule) {
        return ruleRepository.save(rule);
    }



    public Set<RuleModel> commonLeftMerge(Collection<RuleModel> rules) {
        Map<String, RuleModel> calcMap = new HashMap<>();
        for (RuleModel rule : rules) {
            RuleModel oldValue = calcMap.get(rule.getLeftSymbol());
            if (oldValue == null) {
                calcMap.put(rule.getLeftSymbol(), rule);
            } else {
               throw new ResponseStatusException(HttpStatus.BAD_REQUEST,"multiple same leftsymbols");
            }
        }
        if (calcMap.size() == rules.size()) {
            return new HashSet<>(rules);
        }
        return new HashSet<>(calcMap.values());
    }

    private void checkSameRules(RuleModel rule, Grammar grammar){
        if (getByLeftSide(rule.getLeftSymbol(),grammar.getId())!=null){
            throw new ResponseStatusException(HttpStatus.BAD_REQUEST,"already have leftSymbol rule");
        }
    }
    @Transactional
    public Pair<RuleModel, Boolean> addRule(RuleModel rule, Grammar grammar) {

        checkSameRules(rule,grammar);

        if (grammar.getRootSymbol() != null && rule.getRoot()) {
            throw new ResponseStatusException(HttpStatus.BAD_REQUEST, grammar.getId() + " already have root");
        }
        rule = saveRule(rule);
        if (rule.getRoot()!=null && rule.getRoot()) {
            grammar.setRootSymbol(rule.getLeftSymbol());
        }
        grammar.getRules().add(rule);
        rule.setGrammar(grammar);
        grammar.setUpdateTime(LocalDateTime.now());
        return new Pair<>(rule, true);
    }

    @Transactional
    public void deleteRules() {
        grammarRepository.findAll().forEach(((grammar) -> {
            grammar.setRules(null);
            grammar.setRootSymbol(null);
        }));
        ruleRepository.deleteAll();;
    }

    @Transactional
    public void deleteById(Long id) {
        RuleModel rule = ruleRepository.findById(id).orElseThrow(() -> new ResponseStatusException(HttpStatus.NOT_FOUND, "rule id: " + id + " not found"));
        Grammar rulesGrammar = rule.getGrammar();
        rulesGrammar.getRules().remove(rule);
        if (rule.getLeftSymbol().equals(rulesGrammar.getRootSymbol()) ){
            rulesGrammar.setRootSymbol(null);
        }
        ruleRepository.delete(rule);
    }


    public List<RuleModel> getByLeftSide(String leftSymbol) {
        return ruleRepository.findByLeftSymbol(leftSymbol);
    }

    public RuleModel getByLeftSide(String leftSymbol, Long grammarId) {
        return ruleRepository.findByLeftSymbolAndGrammar_Id(leftSymbol, grammarId);
    }



    public RuleModel getById(Long id) {
        return ruleRepository.findById(id).orElseThrow(() -> new ResponseStatusException(HttpStatus.NOT_FOUND));
    }

    @Transactional
    public RuleModel updateRule(Long id, RuleModel rule) {
        RuleModel oldRule = this.getById(id);
        if (rule.getRoot() != null && oldRule.getRoot() != rule.getRoot()) {
            oldRule = changeRootStatus(id);
        }
        boolean changes = false;
        String leftSymbol = rule.getLeftSymbol();
        if (leftSymbol != null && leftSymbol.trim().length() != 0) {
            leftSymbol = rule.getLeftSymbol();
            if (!Objects.equals(leftSymbol, oldRule.getLeftSymbol())) {
                oldRule.setLeftSymbol(leftSymbol);
                changes = true;
            }
        }
       String rightSide = rule.getRightSide();
        if (rightSide != null && rightSide.length() != 0 ) {
            if (!Objects.equals(rightSide, oldRule.getRightSide())) {
                oldRule.setRightSide(rightSide);
                changes = true;
            }
        }
        if (changes) {
            oldRule.getGrammar().setUpdateTime(LocalDateTime.now());
            checkSameRules(rule,rule.getGrammar());

        }
        return oldRule;
    }

    @Transactional
    public RuleModel changeRootStatus(Long id) {
        RuleModel rule = this.getById(id);
        Grammar grammar = rule.getGrammar();
        if (rule.getRoot()) {
            grammar.setRootSymbol(null);
            rule.setRoot(false);
            return rule;
        }
        if (grammar.getRootSymbol() != null) {
            throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "grammar have root terminal");
        }
        grammar.setRootSymbol(rule.getLeftSymbol());
        rule.setRoot(true);
        return rule;

    }


    public Pair<Set<RuleModel>, Boolean> addRuleMultiple(Collection<RuleModel> rules, Grammar grammar) {
        rules = commonLeftMerge(rules);
        Boolean change = false;
        Set<RuleModel> result = new HashSet<>();
        for (RuleModel rule : rules) {
            Pair<RuleModel, Boolean> partResult = addRule(rule, grammar);
            change |= partResult.getValue1();
            result.add(partResult.getValue0());
        }
        return new Pair<>(result, change);

    }

    public RuleModel getRuleByRuleIdGrammarID(long ruleId, long grammarId) {
        return ruleRepository.findByIdAndGrammar_Id(ruleId, grammarId).orElseThrow(() -> new ResponseStatusException(HttpStatus.NOT_FOUND, "rule " + ruleId + " not found in" + grammarId));
    }
}
