package com.jakubu9333.grammarManager;


import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.RuleModel;
import com.jakubu9333.grammarManager.service.GrammarService;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;

import javax.transaction.Transactional;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.Set;



/**
 * @author Jakub Uhlarik
 */


@SpringBootTest
public class GrammarServiceTest {
    @Autowired
    private GrammarService grammarService;


    @AfterEach
    public void tearDown() {
        grammarService.deleteAll();
    }

    @Test
    public void testAddBasicGrammars() {
        Grammar grammarEmpty = new Grammar();
        Set<RuleModel> oldrules = new HashSet<>();
        oldrules.add(new RuleModel(null, "Xd", true));
        grammarEmpty = grammarService.processAddGrammar(grammarEmpty);

        assert null != grammarEmpty.getId();
        assert grammarEmpty.getRules() != null;
        assert grammarEmpty.getUri().equals("null");

    }

    @Test
    public void testSameUriGrammar() {
        Grammar grammar1 = new Grammar("demoUri", null, null);
        Grammar grammar2 = new Grammar("demoUri", null, null);
        grammar1 = grammarService.processAddGrammar(grammar1);
        grammar2 = grammarService.processAddGrammar(grammar2);

        assert grammar1.getUri().equals("demoUri");
        assert grammar2.getUri().matches("demoUri[0-9]+");
        Long id = grammar2.getId();

        Grammar grammar3 = grammarService.processAddGrammar(new Grammar("demoUri" + (id + 2), null, null));
        assert grammar3.getId() == id + 1;
        Grammar grammar4 = grammarService.processAddGrammar(new Grammar("demoUri", null, null));
        assert grammar4.getId() == id + 2;
        assert grammar4.getUri().matches("demoUri[0-9]+.[0-9]+");
    }

    @Test
    public void testMoreThan256() {
        String largeUri = """
                 uri1111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111
                uri1111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111
                uri1111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111
                 """;
        grammarService.processAddGrammar(new Grammar(largeUri));
    }




    @Test
    public void testRootRuleWithRules() {
        Set<RuleModel> rulesRootRule = new HashSet<>();
        RuleModel rule1 = new RuleModel(null, "<p>", ("pp"), true);
        RuleModel rule2 = new RuleModel(null, "<g>", ("pp"), false);
        RuleModel rule3 = new RuleModel(null, "<s>", ("pp"), false);
        rulesRootRule.add(rule1);
        rulesRootRule.add(rule2);
        rulesRootRule.add(rule3);
        boolean exp = false;
        Grammar rootRuleGrammar = new Grammar("root", rulesRootRule, rule2.getLeftSymbol());

            grammarService.processAddGrammar(rootRuleGrammar);

    }


    @Test
    public void testAllGetDelete() {
        Set<RuleModel> ruleSet = new HashSet<>();
        RuleModel rule1 = new RuleModel(null, "rule1", "smile", false);
        RuleModel rule2 = new RuleModel(null, "rule2", "cry|sis", false);
        ruleSet.add(rule1);
        ruleSet.add(rule2);
        Grammar grammar1 = new Grammar("1", ruleSet, null);
        Grammar grammar2 = new Grammar("2", null, null);
        Grammar grammar3 = new Grammar("3", null, null);
        Grammar grammar4 = new Grammar("4", null, null);
        grammarService.processAddGrammar(grammar1);
        grammarService.processAddGrammar(grammar2);
        grammarService.processAddGrammar(grammar3);
        grammarService.processAddGrammar(grammar4);
        List<Grammar> grammars = grammarService.getGrammars();
        assert grammars.size() == 4;
        int one = 0;
        int two = 0;
        int three = 0;
        int four = 0;
        int five = 0;
        for (Grammar iterGram : grammars) {
            switch (iterGram.getUri()) {
                case "1" -> one++;
                case "2" -> two++;
                case "3" -> three++;
                case "4" -> four++;
                default -> five++;
            }
        }
        assert one == 1 && two == 1 && three == 1 && four == 1 && five == 0;
        grammarService.deleteAll();
        List<Grammar> grammars2 = grammarService.getGrammars();
        assert grammars2.size() == 0;
    }

    @Test
    public void testChangeUri() {
        Grammar grammar1 = new Grammar("1x", null, null);
        Grammar grammar2 = new Grammar("2x", null, null);
        grammar1 = grammarService.processAddGrammar(grammar1);
        grammar2 = grammarService.processAddGrammar(grammar2);
        Long id = grammar2.getId();
        Grammar changeGr = new Grammar("xd", null, null);
        grammar2 = grammarService.changeUri(id, changeGr);
        assert Objects.equals(grammar2.getId(), id);
        assert grammar2.getUri().equals("xd");
        grammar1 = grammarService.changeUri(grammar1.getId(), changeGr);
        assert grammar1.getUri().equals("xd" + grammar1.getId());
        Grammar grammar3 = new Grammar("3x", null, null);
        Grammar changeGr2;
        grammar3 = grammarService.processAddGrammar(grammar3);
        changeGr2 = grammarService.changeUri(grammar3.getId(), new Grammar(null, null, null));
        assert changeGr2.getUri().equals("3x");
        changeGr2 = grammarService.changeUri(grammar3.getId(), new Grammar("", null, null));
        assert changeGr2.getUri().equals("3x");
        changeGr2 = grammarService.changeUri(grammar3.getId(), new Grammar("    \n   " +
                "           ", null, null));
        assert changeGr2.getUri().equals("3x");
        changeGr2 = grammarService.changeUri(grammar3.getId(), new Grammar("  ", null, null));
        assert changeGr2.getUri().equals("3x");
    }

    @Test
    @Transactional
    public void testDuplicate() {
        Set<RuleModel> ruleSet = new HashSet<>();
        RuleModel rule1 = new RuleModel(null, "rule1","smile", false);
        RuleModel rule = new RuleModel(null, "rule1", "smile", false);
        RuleModel rule2 = new RuleModel(null, "rule2", "cry|sis", false);
        ruleSet.add(rule1);
        ruleSet.add(rule2);
        Grammar grammar1 = new Grammar("1x", null, rule.getLeftSymbol());
        Grammar grammar2 = new Grammar("2x", ruleSet, null);
        grammar1 = grammarService.processAddGrammar(grammar1);
        grammar2 = grammarService.processAddGrammar(grammar2);
        grammarService.duplicateGrammar(grammar1.getId());
        grammarService.getGrammars();
        grammarService.duplicateGrammar(grammar2.getId());
        grammarService.getGrammars();
    }

    @Test
    public void testGetIdDeleteId() {
        Set<RuleModel> ruleSet = new HashSet<>();
        RuleModel rule1 = new RuleModel("rule1", "smile", false);
        RuleModel rule2 = new RuleModel("rule2", "cry|sis", false);
        ruleSet.add(rule1);
        ruleSet.add(rule2);
        Grammar grammar1 = new Grammar("1x", null, null);
        Grammar grammar2 = new Grammar("2x", ruleSet, null);
        Grammar grammar3 = new Grammar("3x", null, null);
        Grammar grammar4 = new Grammar("4x", null, null);
        grammar1 = grammarService.processAddGrammar(grammar1);
        grammar2 = grammarService.processAddGrammar(grammar2);
        grammarService.processAddGrammar(grammar3);
        grammarService.processAddGrammar(grammar4);
        assert grammarService.getGrammarById(grammar1.getId()).getUri().equals("1x");
        grammarService.deleteById(grammar2.getId());
        List<Grammar> grammars = grammarService.getGrammars();
        assert grammars.size() == 3;
        int one = 0;
        int two = 0;
        int three = 0;
        int four = 0;
        int five = 0;
        for (Grammar iterGram : grammars) {
            switch (iterGram.getUri()) {
                case "1x" -> one++;
                case "2x" -> two++;
                case "3x" -> three++;
                case "4x" -> four++;
                default -> five++;
            }
        }
        assert one == 1 && two == 0 && three == 1 && four == 1 && five == 0;
    }



    @Test
    @Transactional
    public void testUriStartUri() {
        Grammar grammar = new Grammar("go1", null, null);
        Grammar grammar2 = new Grammar("go2", null, null);
        Grammar grammar3 = new Grammar("gl3", null, null);
        Grammar grammar4 = new Grammar("gofsdfjdsfpsdfosddslfksflk3", null, null);
        grammarService.processAddGrammar(grammar);
        grammarService.processAddGrammar(grammar2);
        grammarService.processAddGrammar(grammar3);
        grammarService.processAddGrammar(grammar4);
        List<Grammar> grammars = grammarService.getGrammarsByUriStart("go");
        assert grammars.size() == 3;
        for (Grammar iterGram : grammars) {
            assert iterGram.getUri().startsWith("go");
        }
        assert grammarService.getGrammarsByUriStart("fsdfsdf").size() == 0;
    }


}
