package com.jakubu9333.grammarManager;

import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.RuleModel;
import com.jakubu9333.grammarManager.service.GrammarService;
import com.jakubu9333.grammarManager.service.RuleService;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.transaction.annotation.Transactional;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * @author Jakub Uhlarik
 */
@SpringBootTest
public class RuleServiceTest {

    @Autowired
    private RuleService ruleService;
    @Autowired
    private GrammarService grammarService;


    @AfterEach
    public void tearDown() {
        grammarService.deleteAll();
    }


    @Test
    public void basicRulesAdd() {
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        RuleModel rule = new RuleModel(null, "h",("xd"), false);
        RuleModel rule1 = new RuleModel(null, "g",("xd"), true);
        rule = ruleService.addRule(rule, grammar).getValue0();
        ruleService.addRule(rule1, grammar);
        assert grammar.getRootSymbol().equals("g");
        assert rule.getId() != null;
        Set<RuleModel> rules = grammar.getRules();
        assert rules.size() == 2;
        int one = 0;
        int two = 0;
        int three = 0;
        for (RuleModel smallRule : rules) {
            switch (smallRule.getLeftSymbol()) {
                case "h" -> one++;
                case "g" -> two++;
                default -> three++;
            }
        }
        assert one == 1 && two == 1 && three == 0;


    }




    @Test
    public void addMultipleRuleTest() {
        RuleModel rule2 = new RuleModel(null, "<rule2>", "ddggd", true);
        RuleModel rule3 = new RuleModel(null, "<rule3>",("smiling"), false);
        RuleModel rule4 = new RuleModel(null, "<rule4>","xdd", false);
        RuleModel rule5 = new RuleModel(null, "<rule5>","xdgd", false);
        Grammar grammar = grammarService.addGrammar(new Grammar());
        Set<RuleModel> rules = new HashSet<>(List.of(rule4, rule2, rule3, rule5));
        ruleService.addRuleMultiple(rules, grammar);
        assert grammar.getRules().equals(rules);

        assert !ruleService.addRuleMultiple(new HashSet<>(), grammar).getValue1();
        assert grammar.getRules().equals(rules);

        Grammar grammar1 = grammarService.addGrammar(new Grammar());
        ruleService.addRuleMultiple(new HashSet<>(), grammar1);
        assert grammar1.getRules().size() == 0;
        RuleModel rule6 = new RuleModel(null, "rule2",("(d|g)dggd"), true);
        RuleModel rule7 = new RuleModel(null, "rule3",("smiling"), false);
        Set<RuleModel> rules2 = new HashSet<>(List.of(rule6, rule7));
        Grammar grammar3 = grammarService.addGrammar(new Grammar());
        ruleService.addRuleMultiple(rules2, grammar3);
        RuleModel rule8 = new RuleModel(null, "rule2","(d|g)dggd", true);
        assert grammar3.getRules().equals(new HashSet<>(List.of(rule7, rule8)));
    }

    @Test
    public void testEmptyAdd() {
        Grammar gr = new Grammar();
        grammarService.addGrammar(gr);
        RuleModel nullRule = new RuleModel();
        RuleModel nullRightStr = new RuleModel(null, "dd", null, null);
        RuleModel emptyStrRule = new RuleModel(null, "",(""), null);
         ruleService.addRule(nullRule,gr);
        emptyStrRule = ruleService.addRule(emptyStrRule, gr).getValue0();
        nullRightStr = ruleService.addRule(nullRightStr, gr).getValue0();
        assert emptyStrRule.getRightSide().equals((""));
        assert nullRightStr.getRightSide()==null;
    }










    @Test
    public void testListDeleteAll() {
        ruleService.deleteRules();
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        RuleModel rule = new RuleModel(null, "1",("xd"), false);
        RuleModel rule1 = new RuleModel(null, "2",("d"), true);
        ruleService.addRule(rule, grammar);
        ruleService.addRule(rule1, grammar);
        int one = 0;
        int two = 0;
        int three = 0;
        List<RuleModel> rules = ruleService.getAllRules();
        assert rules.size() == 2;
        for (RuleModel partRule : rules) {
            switch (partRule.getLeftSymbol()) {
                case "1" -> one++;
                case "2" -> two++;
                default -> three++;
            }
        }
        assert one == 1 && two == 1 && three == 0;
        ruleService.deleteRules();
        assert ruleService.getAllRules().size() == 0;
    }




    @Test
    public void testFindDeleteById() {
        ruleService.deleteRules();
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        RuleModel rule = new RuleModel(null, "1",("xd"), false);
        RuleModel rule1 = new RuleModel(null, "2",("d"), true);
        RuleModel rule3 = new RuleModel(null, "3",("d"), false);
        rule = ruleService.addRule(rule, grammar).getValue0();
        rule1 = ruleService.addRule(rule1, grammar).getValue0();
        ruleService.addRule(rule3, grammar);
        assert ruleService.getById(rule.getId()).getLeftSymbol().equals("1");
        ruleService.deleteById(rule1.getId());
        List<RuleModel> rules = ruleService.getAllRules();
        assert rules.size() == 2;
        int one = 0;
        int def = 0;
        int three = 0;
        for (RuleModel smallRule : rules) {
            switch (smallRule.getLeftSymbol()) {
                case "1" -> one++;
                case "3" -> three++;
                default -> def++;
            }
        }
        assert one == 1 && three == 1 && def == 0;


    }

    @Test
    @Transactional
    public void testGetByRightSide() {
        ruleService.deleteRules();
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Grammar grammar2 = new Grammar();
        grammar2 = grammarService.addGrammar(grammar2);
        RuleModel rule = new RuleModel(null, "1",("xd"), false);
        RuleModel rule1 = new RuleModel(null, "2",("d"), true);
        RuleModel rule3 = new RuleModel(null, "1",("d"), false);
        ruleService.addRule(rule, grammar);
        ruleService.addRule(rule1, grammar);
        ruleService.addRule(rule3, grammar2);
        List<RuleModel> rules = ruleService.getByLeftSide("1");
        assert rules.size() == 2;
        for (RuleModel smallRule : rules) {
            assert smallRule.getLeftSymbol().equals("1");
        }
        RuleModel rulesFromGrammar = ruleService.getByLeftSide("1", grammar.getId());
        assert rulesFromGrammar!= null;

            assert rulesFromGrammar.getLeftSymbol().equals("1") && rulesFromGrammar.getRightSide().equals(("xd"));

        assert ruleService.getByLeftSide("5").size() == 0;
    }
}
